/**
 * The class that implements the SortedSet spec.
 * 
 * @param <E> the type of the TreeSet
 * 
 * @author Gloss
 * @version 1.1 - 11/Jul/2013 asantos
 **/
public class TreeSet<E extends IOrderable<E>> implements Cloneable {

	/** The root node of the tree **/
	private Node<E> root;

	/**
	 * Class that implements a TreeSet node.
	 * 
	 * @param <E> - the type of the node.
	 * 
	 * @author Gloss
	 **/
	static class Node<E extends IOrderable<E>> {

		/** The right and left nodes **/
		Node<E> left, right;

		/** The next node wrt increasing order; only calculated/updated when needed **/
		Node<E> next;

		/** The node data **/
		E data;

		/**
		 * Create a new node.
		 * 
		 * @param left - the left node of the current node.
		 * @param right - the right node of the current node.
		 * @param data - the current node value.
		 **/
		Node(Node<E> left, E data, Node<E> right) {
			this.left = left;
			this.right = right;
			this.data = data;
		}

		/**
		 * The node representation.
		 * 
		 * @return the representation of the node.
		 **/
		@Override
		public String toString() {
			return data.toString();
		}

		/**
		 * Hashcode for Node
		 * 
		 * @return hashcode for the Node
		 */
		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((data == null) ? 0 : data.hashCode());
			result = prime * result + ((left == null) ? 0 : left.hashCode());
			result = prime * result + ((next == null) ? 0 : next.hashCode());
			result = prime * result + ((right == null) ? 0 : right.hashCode());
			return result;
		}

		/**
		 * Compares if the object is equal to Node.
		 * 
		 * @param obj - the object to be compared
		 * 
		 * @return true if they are equal, otherwise return false.
		 **/
		@SuppressWarnings("unchecked")
		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (!(obj instanceof Node<?>))
				return false;
			Node<E> other = (Node<E>) obj;
			return equalElements(other);
		}

		/**
		 * Auxiliary method to compare data and connections of the node
		 * @param other The node to compare to this
		 * @return true if the date and connections are the equal, false otherwise
		 */
		private boolean equalElements(Node<E> other) {

			if (data == null) {
				if (other.data != null)
					return false;
			} else if (!data.equals(other.data))
				return false;

			if (left == null) {
				if (other.left != null)
					return false;
			} else if (!left.equals(other.left))
				return false;

			if (next == null) {
				if (other.next != null)
					return false;
			} else if (!next.equals(other.next))
				return false;

			if (right == null) {
				if (other.right != null)
					return false;
			} else if (!right.equals(other.right))
				return false;

			return true;
		}

		/**
		 * Makes a copy of the Node
		 * @return copy of this Node
		 */
		public Node<E> copy(){

			Node<E> rightResult = null;
			Node<E> leftResult = null;
			if (right != null)
				rightResult = right.copy();
			if (left != null)
				leftResult = left.copy();
			return new Node<E>(leftResult, data, rightResult);
		}
	} //end Node

	/**
	 * The default constructor for TreeSet.
	 **/
	public TreeSet() {
		root = null;
	}

	/**
	 * Insert the new node in the tree, relatively to the given node.
	 * 
	 * @param node - the relative node to the new node.
	 * @param e - the value of the node that will be inserted in the tree.
	 * @return the current node.
	 **/
	private Node<E> insert(Node<E> node, E e) {

		if (node == null)
			return new Node<E>(null, e, null);
		if (e.equals(node.data))
			return node;

		if (e.isGreaterEq(node.data))
			node.right = insert(node.right, e);
		else
			node.left = insert(node.left, e);
		return node;
	}

	/**
	 * Insert the new node relatively to the root node of the tree.
	 * 
	 * @param e - the value of the node that will be inserted in the tree.
	 **/
	public void insert(E e) {
		root = insert(root, e);
	}

	/**
	 * Verifies if the tree is empty.
	 * 
	 * @return true if the tree is empty, false otherwise
	 **/
	public boolean isEmpty() {
		return root == null;
	}

	/**
	 * Verifies if a node with the given value exists in the tree. 
	 * Starts the search from the given node.
	 * 
	 * @param node - the node where the search begins
	 * @param e - the value to be found.
	 * @return true if a node with the given value exists in the tree, 
	 *         false otherwise
	 **/
	private boolean isIn(Node<E> node, E e) {
		if (node == null)
			return false;
		if (node.data.equals(e))
			return true;
		return e.isGreaterEq(node.data)? isIn(node.right, e) : isIn(node.left, e);
	}

	/**
	 * Verifies if a node with the given value exists in the tree.
	 * Starts the search from the root node of the tree.
	 * 
	 * @param e - the value to be found.
	 * @return true if a node with the given value exists in the tree, 
	 *         false otherwise
	 **/
	public boolean isIn(E e) {
		return isIn(root, e);
	}

	/**
	 * Get the largest value in the tree.
	 * 
	 * @return the largest value in the tree.
	 */
	public E largest() {
		if (root == null)
			return null;
		Node<E> rightMost = root;
		while (rightMost.right != null)
			rightMost = rightMost.right;
		return rightMost.data;
	}

	/**
	 * Creates a copy of the current tree.
	 * 
	 * @return the clone of the current set.
	 **/
	@SuppressWarnings("unchecked")
	@Override
	public TreeSet<E> clone() {
		try {
			TreeSet<E> result = (TreeSet<E>) super.clone();
			if (root != null)
				result.root = root.copy();	
			return result;
		} catch (CloneNotSupportedException e) {
			throw new InternalError(e.getMessage());
		}
	}

	/**
	 * Hashcode for TreeSet
	 * 
	 * @return hashcode for the TreeSet
	 */
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((root == null) ? 0 : root.hashCode());
		return result;
	}

	/**
	 * Compares this tree to another object to test if they are equal.
	 * 
	 * @param other - the reference object with which to compare.
	 * @return true if object is equal, false otherwise.
	 **/
	@SuppressWarnings("unchecked")
	@Override
	public boolean equals(Object other) {
		return this == other || 
				((other instanceof TreeSet) && equalTo((TreeSet<E>) other));
	}

	/**
	 * Verifies if the given tree is equal to the current.
	 * 
	 * @param ts - the TreeSet to compare with.
	 * @return true if they have the same elements in the same position
	 **/
	private boolean equalTo(TreeSet<E> ts) {
		boolean equals = true;
		
		Node<E> node1 = visit(root, null);
		Node<E> node2 = ts.visit(ts.root, null);

		while (node1 != null && node2 != null && equals){
			equals = node1.data.equals(node2.data);
			node1 = node1.next;
			node2 = node2.next;
		}
		
		equals =equals && (node1 == null && node2 == null);
		return equals;
	}

	/**
	 * Visit the sub-tree from the given node, connecting
	 * the nodes during the visit in the increasing order
	 * Returns the left most node, i.e., the smallest.
	 * 
	 * @param node - the root node of the sub-tree.
	 * @return the leftmost node
	 **/
	private Node<E> visit(Node<E> node, Node<E> next) {
		if (node == null)
			return null;

		Node<E> leftMost;
		if (node.left != null)
			leftMost = visit(node.left, node);
		else
			leftMost = node;
		if (node.right != null)
			node.next = visit(node.right, next);
		else
			node.next = next;
		return leftMost;
	}

	/**
	 * The tree textual representation.
	 * 
	 * @return the String representing the tree
	 */
	@Override
	public String toString() {
		StringBuilder tree = new StringBuilder("{");
		if(root != null) {
			printTree(root, tree);
			return tree.toString()+"}";
		}
		return "{ }";
	}

	/**
	 * Builds the tree String representation
	 * @param node - the current node
	 * @param tree - the Stringbuilder being built
	 */
	private void printTree(Node<E> node, StringBuilder tree) {
		if(node != null) {
			tree.append(node.data.toString() + " ");
			printTree(node.left, tree);
			printTree(node.right, tree);
		}
	}
}
